import hashlib
import json
import traceback
from worker import WorkerExit, sleep

from .save_path import SavePath
from .module_grabber import ModuleGrabber
from .image import Image
from .error import LastPageError, PauseDownloadError, SkipEpisodeError, is_http, SkipPageError, ComicCrawlerError

from .io import path_each, content_read, content_write
from .util import url_extract_filename, debug_log
from .channel import download_ch, mission_ch
from .safeprint import print

VALID_FILE_TYPES = (
	# images
	".jpg", ".jpeg", ".gif", ".png", ".svg", ".psd", ".webp", ".bmp",
	# zips
	".zip", ".rar",
	# videos
	".mp4", ".m4v", ".mkv", ".swf", ".webm", ".mov", ".wmv",
	# audio
	".mp3", ".aac", ".flac", ".wav",
	# json
	".json", ".txt"
)

class Crawler:
	"""Create Crawler object. Contains img url, next page url."""
	def __init__(self, mission, ep, savepath):
		"""Construct."""
		self.mission = mission
		self.ep = ep
		self.savepath = SavePath(savepath, mission, ep)
		self.mod = mission.module
		self.downloader = ModuleGrabber(mission.module)
		self.checksums = None
		self.is_init = False
		self.html = None
		self.image = None
		self.images = None
		self.image_bin = None
		self.image_ext = None
		self.filename = None
		
	def init(self):
		if not self.ep.current_url:
			self.ep.current_url = self.ep.url
		if not self.ep.current_page:
			self.ep.current_page = 1
			
		self.init_images(self.ep.current_page - 1)
			
		self.is_init = True
		
	def init_images(self, skip_pages=0):
		"""Grab html, images and move cursor to correct image by skip_pages"""
		self.get_html()
		self.get_images()
		
		try:
			# skip some images
			for _ in range(0, skip_pages):
				next(self.images)
			# get current image
			self.image = Image.create(next(self.images))
		except StopIteration:
			self.image = None
			
	def get_filename(self):
		if self.mission.module.config.getboolean("originalfilename"):
			return self.image.filename
		return self.ep.total + 1
		
	def page_exists(self):
		return self.savepath.exists(self.get_filename())
			
	def download_image(self):
		"""Download image"""
		if self.image.url:
			result = self.downloader.img(
				self.image.url,
				referer=None if getattr(self.mod, "no_referer", False) else self.ep.current_url
			)
				
			if result.response.history:
				self.handle_redirect(result.response)
				
			# redirected and url changed
			if result.response.history and not self.image.static_filename:
				self.image.filename = url_extract_filename(result.response.url)
			bin = result.bin
			ext = result.ext
		else:
			bin = json.dumps(self.image.data, indent="\t").encode("utf-8")
			ext = ".json"
			
		if not ext:
			raise ValueError("Bad file type, the file extension is undefined")
			
		if ext not in VALID_FILE_TYPES:
			raise ValueError("Bad file type: " + ext)
			
		self.image_bin = bin
		self.image_ext = ext
		
	def handle_redirect(self, response):
		"""FIXME: should we merge this into handle_image?"""
		if hasattr(self.mod, "redirecthandler"):
			self.mod.redirecthandler(response, self)
			
	def handle_image(self):
		"""Post processing"""
		if hasattr(self.mod, "imagehandler"):
			self.image_ext, self.image_bin = self.mod.imagehandler(
				self.image_ext, self.image_bin)

	def save_image(self):
		"""Write image to save path"""
		if getattr(self.mod, "circular", False):
			if not self.checksums:
				self.checksums = set()
				path_each(
					self.savepath.parent(),
					lambda file: self.checksums.add(get_file_checksum(file))
				)

			checksum = get_checksum(self.image_bin)
			if checksum in self.checksums:
				raise LastPageError
			self.checksums.add(checksum)
				
		try:
			content_write(self.savepath.full_fn(self.get_filename(), self.image_ext), self.image_bin)
		except OSError as err:
			traceback.print_exc()
			raise PauseDownloadError("Failed to write file!") from err

	def next_page(self):
		"""Iter to next page."""
		next_page = self.get_next_page()
		if not next_page:
			raise LastPageError
			
		self.ep.current_url = next_page
		self.ep.current_page = 1
		
		self.init_images()

	def next_image(self):
		self.ep.current_page += 1
		self.ep.total += 1
		try:
			self.image = Image.create(next(self.images))
		except StopIteration:
			self.image = None
			
	def resolve_image(self):
		self.image.resolve()
		
	def rest(self):
		"""Rest some time."""
		sleep(getattr(self.mod, "rest", 0))
		
	def get_next_page(self):
		# self.html is not a str if self.ep.image is not None
		if not isinstance(self.html, str):
			return
		if hasattr(self.mod, "get_next_image_page"):
			return self.mod.get_next_image_page(self.html, self.ep.current_url)
		if hasattr(self.mod, "get_next_page"):
			return self.mod.get_next_page(
				self.html,
				self.ep.current_url
			)
		
	def get_html(self):
		if self.ep.image:
			self.html = True
		else:
			self.html = self.downloader.html(self.ep.current_url, referer=self.mission.url)
		
	def get_images(self):
		"""Get images"""
		skip_page = False
		if self.ep.image:
			images = self.ep.image
		else:
			try:
				images = self.mod.get_images(
					self.html,
					self.ep.current_url
				)
			except SkipPageError:
				images = []
				skip_page = True
		if isinstance(images, str):
			images = [images]
		if not images and not skip_page:
			raise ValueError("get_images returns an empty array")
		try:
			self.images = iter(images)
		except TypeError:
			self.images = iter([images])
		
	def handle_error(self, error):
		"""Send error to error handler."""
		handler = getattr(self.mod, "errorhandler", None)
		if not handler:
			return

		try:
			handler(error, self)

		except Exception as err: # pylint: disable=broad-except
			print("[Crawler] Failed to handle error: {}".format(err))
			
def get_checksum(b):
	return hashlib.md5(b).hexdigest() # nosec

def get_file_checksum(file):
	return get_checksum(content_read(file, raw=True))

def download(mission, savepath):
	"""Download mission to savepath."""

	# warning there is a deadlock,
	# never do mission.lock.acquire in callback...
	print("Start downloading " + mission.title)
	mission.state = "DOWNLOADING"
	try:
		crawl(mission, savepath)

		# Check if mission is complete
		for ep in mission.episodes:
			if not ep.complete and not ep.skip:
				raise ComicCrawlerError("Mission is not completed")

	except WorkerExit:
		mission.state = "PAUSE"
		download_ch.pub('DOWNLOAD_PAUSE', mission)
		raise

	except PauseDownloadError as err:
		mission.state = "ERROR"
		download_ch.pub('DOWNLOAD_INVALID', (err, mission))

	except Exception as err:
		mission.state = "ERROR"
		download_ch.pub('DOWNLOAD_ERROR', (err, mission))
		raise

	else:
		mission.state = "FINISHED"
		download_ch.pub("DOWNLOAD_FINISHED", mission)	
	
def crawl(mission, savepath):
	"""Crawl each episode."""
	
	print("total {} episode.".format(len(mission.episodes)))

	for ep in mission.episodes:
		if ep.skip or ep.complete:
			continue

		print("Downloading ep {}".format(ep.title))
		
		try:
			crawler = Crawler(mission, ep, savepath)
			crawlpage(crawler)

		except LastPageError:
			print("Episode download complete!")
			ep.complete = True
			download_ch.pub("DOWNLOAD_EP_COMPLETE", (mission, ep))

		except SkipEpisodeError as err:
			print("Something bad happened, skip the episode.")
			if err.always:
				ep.skip = True

def crawlpage(crawler):
	"""Crawl all pages of an episode.

	To complete current episode, raise LastPageError.
	To skip current episode, raise SkipEpisodeError.
	To stop downloading (fatal error), raise PauseDownloadError.
	"""
	
	def download():
		if not crawler.is_init:
			debug_log("D_INIT")
			crawler.init()
			
		if not crawler.html:
			debug_log("D_INIT_IMAGE")
			crawler.init_images()
	
		if not crawler.image:
			debug_log("D_NEXT_PAGE")
			crawler.next_page()
			return
			
		if crawler.page_exists():
			debug_log("D_NEXT_IMAGE")
			print("page {} already exist".format(crawler.ep.total + 1))
			crawler.next_image()
			return
			
		debug_log("D_RESOLVE")
		crawler.resolve_image()
		print("Downloading {} page {}: {}\n".format(
			crawler.ep.title, crawler.ep.total + 1, crawler.image.url))
		debug_log("D_DOWNLOAD")
		crawler.download_image()
		debug_log("D_HANDLE")
		crawler.handle_image()
		debug_log("D_SAVE")
		crawler.save_image()
		debug_log("D_PUB")
		mission_ch.pub("MISSION_PROPERTY_CHANGED", crawler.mission)
		debug_log("D_REST")
		crawler.rest()
		debug_log("D_NEXT_IMAGE")
		crawler.next_image()

	def download_error(er, count):
		t = 5 * 2 ** count
		print(f"wait {t} seconds...")
		if is_http(er, code=429):
			# retry doesn't work with 429 error
			sleep(t)
			raise er
		crawler.handle_error(er)
		sleep(t)

	error_loop(download, download_error)

def error_loop(process, handle_error=None, limit=3):
	"""Loop process until error. Has handle error limit."""
	errorcount = 0
	while True:
		try:
			process()
		except Exception as er: # pylint: disable=broad-except
			traceback.print_exc()
			errorcount += 1
			if errorcount >= limit:
				raise SkipEpisodeError(always=False) from None
			if handle_error:
				handle_error(er, errorcount)
		# except ExitErrorLoop:
			# break
		else:
			errorcount = 0



#! python3

"""this is pixiv module for comiccrawler

Ex:
	http://www.pixiv.net/member_illust.php?id=2211832

"""

import re
import json
from html import unescape
from io import BytesIO
from urllib.parse import urljoin, urlencode, urlparse, parse_qs
from zipfile import ZipFile

from ..core import Episode, grabhtml
from ..error import PauseDownloadError, is_http, SkipEpisodeError, SkipPageError
from ..safeprint import print

domain = ["www.pixiv.net"]
name = "Pixiv"
noepfolder = True
config = {
	"cookie_PHPSESSID": "請輸入Cookie中的PHPSESSID"
}

class DataNotFound(Exception):
	pass

def get_init_data(html):
	match = re.search("""meta-global-data" content='([^']+)""", html)
	if not match:
		raise DataNotFound
	data = json.loads(unescape(match.group(1)))
		
	match = re.search("""meta-preload-data" content='([^']+)""", html)
	# if not match:
		# raise DataNotFound
	data["preload"] = json.loads(unescape(match.group(1)))
	
	return data

def get_title_from_init_data(html, url):
	init_data = get_init_data(html)
	user = next(iter(init_data["preload"]["user"].values()))
	tag = get_tag_from_url(url)
	tag = " ({})".format(tag) if tag else ""
	return "{} - {}{}".format(user["userId"], user["name"], tag)
	
def is_search_page(url):
	return re.match("https://www\.pixiv\.net/tags/", url)

def get_title(html, url):
	if is_search_page(url):
		# general title?
		pass
	else:
		try:
			return get_title_from_init_data(html, url)
		except DataNotFound:
			pass
	return "[pixiv] " + unescape(re.search("<title>([^<]+)", html).group(1))
	
def check_login(data):
	if not data.get("userData"):
		raise PauseDownloadError("you didn't login!")
		
def check_login_html(html):
	if "pixiv.user.loggedIn = true" not in html and "login: 'yes'" not in html:
		raise PauseDownloadError("you didn't login!")
		
cache_next_page = {}

def get_episodes_from_works(works):
	s = []
	for data in sorted(works, key=lambda i: int(i["id"])):
		s.append(Episode(
			"{} - {}".format(data["id"], data["title"]),
			"https://www.pixiv.net/member_illust.php?mode=medium&illust_id={}".format(data["id"])
		))
	return s

def get_tag_from_url(url):
	tags = parse_qs(urlparse(url).query).get("tag")
	return tags[0] if tags else None
	
def get_episodes_from_init_data(html, url):
	init_data = get_init_data(html)
	check_login(init_data)
	
	id = int(next(iter(init_data["preload"]["user"])))
	tag = get_tag_from_url(url)
	
	if tag:
		def build_url(offset):
			query = {
				"offset": str(offset),
				"limit": "48",
				"tag": tag
			}
			return "https://www.pixiv.net/ajax/user/{}/illustmanga/tag?{}".format(id, urlencode(query))
			
		response = grabhtml(build_url(0))
		response = json.loads(response)
		total = response["body"]["total"]
		i = 48
		pre_url = url
		while i < total:
			next_url = build_url(i)
			cache_next_page[pre_url] = next_url
			pre_url = next_url
			i += 48
		return get_episodes_from_works(response["body"]["works"])
	
	all = grabhtml("https://www.pixiv.net/ajax/user/{}/profile/all".format(id))
	all = json.loads(all)
	
	ep_ids = [int(id) for id in list(all["body"]["illusts"]) + list(all["body"]["manga"])]
	ep_ids.sort()
	ep_ids.reverse()
	
	pre_url = url
	for page, i in enumerate(range(0, len(ep_ids), 48)):
		ids = ep_ids[i:i + 48]
		query = [("ids[]", str(id)) for id in ids] + [
			("is_manga_top", "0"),
			("work_category", "illustManga"),
			("is_first_page", "1" if page == 0 else "0")
		]
		new_url = "https://www.pixiv.net/ajax/user/{}/profile/illusts?{}".format(
			id, urlencode(query))
		cache_next_page[pre_url] = new_url
		pre_url = new_url
	raise SkipPageError
	
def get_episodes_from_ajax_result(html, url):
	if "ajax/user" not in url:
		raise DataNotFound
	works = json.loads(html)["body"]["works"]
	if isinstance(works, dict):
		works = works.values()
	return get_episodes_from_works(works)
	
def get_episodes_from_search(html, url):
	word = re.search("/tags/([^/]+)", url).group(1)
	query = urlparse(url).query
	# FIXME: is it safe to reuse the query?
	ajax_url = "https://www.pixiv.net/ajax/search/artworks/{}?{}".format(word, query)
	cache_next_page[url] = ajax_url
	raise SkipPageError
	
def is_search_ajax(url):
	return url.startswith("https://www.pixiv.net/ajax/search/artworks/")
	
def get_episodes_from_search_ajax(html, url):
	data = json.loads(html)
	episodes = [
		Episode(
			"{} - {}".format(i["id"], i["title"]),
			"https://www.pixiv.net/artworks/{}".format(i["id"])
		) for i in data["body"]["illustManga"]["data"]
	]
	
	if episodes:
		url_o = urlparse(url)
		query = parse_qs(url_o.query)
		p = query.get("p", ["1"])[0]
		query["p"] = [str(int(p) + 1)]
		cache_next_page[url] = url_o._replace(query=urlencode(query, doseq=True)).geturl()
	
	return episodes[::-1]
		
def get_episodes(html, url):
	if is_search_page(url):
		return get_episodes_from_search(html, url)
		
	if is_search_ajax(url):
		return get_episodes_from_search_ajax(html, url)

	try:
		return get_episodes_from_ajax_result(html, url)
	except DataNotFound:
		pass
		
	try:
		return get_episodes_from_init_data(html, url)
	except DataNotFound:
		pass

	check_login_html(html)
	s = []
	# search result?
	match = re.search('id="js-mount-point-search-result-list"data-items="([^"]+)', html)
	if match:
		data = unescape(match.group(1))
		for illust in json.loads(data):
			s.append(Episode(
				"{illustId} - {illustTitle}".format_map(illust),
				urljoin(url, "/member_illust.php?mode=medium&illust_id={illustId}".format_map(illust))
			))
			
	# single image
	if "member_illust.php?mode=medium&illust_id" in url:
		s.append(Episode("image", url))
		
	return s[::-1]
	
cache = {}

def get_nth_img(url, i):
	return re.sub(r"_p0(\.\w+)$", r"_p{}\1".format(i), url)

def get_images(html, url):
	if "&amp;" in url:
		# fix bad URL for old saves e.g. 
		# https://www.pixiv.net/member_illust.php?mode=medium&amp;illust_id=12345
		cache_next_page[url] = unescape(url)
		raise SkipPageError

	init_data = get_init_data(html)
	check_login(init_data)
	if len(init_data["preload"]["illust"]) == 1:
		illust_id, illust = init_data["preload"]["illust"].popitem()
	else:
		illust_id = re.search("illust_id=(\d+)", url).group(1)
		illust = init_data["preload"]["illust"][illust_id]
	
	if illust["illustType"] != 2: # normal images
		first_img = illust["urls"]["original"]
		return [get_nth_img(first_img, i) for i in range(illust["pageCount"])]
		
	# https://www.pixiv.net/member_illust.php?mode=medium&illust_id=44298524
	ugoira_meta = "https://www.pixiv.net/ajax/illust/{}/ugoira_meta".format(illust_id)
	ugoira_meta = json.loads(grabhtml(ugoira_meta))
	cache["frames"] = ugoira_meta["body"]["frames"]
	return ugoira_meta["body"]["originalSrc"]

def errorhandler(err, crawler):
	if is_http(err, 404):
		url = None
		try:
			url = err.response.url
		except AttributeError:
			pass
		if url and is_ep_url(url):
			# deleted by author?
			# https://www.pixiv.net/member_illust.php?mode=medium&illust_id=68059323
			print("Skip {}: {}".format(err.response.url, 404))
			raise SkipEpisodeError
			
def is_ep_url(url):
	patterns = [
		"member_illust\.php.*illust_id=\d+",
		"/artworks/\d+"
	]
	return any(re.search(p, url) for p in patterns)
			
def imagehandler(ext, bin):
	"""Append index info to ugoku zip"""
	if ext == ".zip":
		bin = pack_ugoira(bin, cache["frames"])
		ext = ".ugoira"
	return ext, bin
	
def pack_ugoira(bin, frames):
	with BytesIO(bin) as imbin:
		with ZipFile(imbin, "a") as zip:
			data = json.dumps({"frames": frames}, separators=(',', ':'))
			zip.writestr("animation.json", data.encode("utf-8"))
		return imbin.getvalue()

def get_next_page(html, url):
	match = re.search("href=\"([^\"]+)\" rel=\"next\"", html)
	if match:
		return urljoin(url, unescape(match.group(1)))
		
	if url in cache_next_page:
		next_url = cache_next_page[url]
		del cache_next_page[url]
		return next_url		



"""
https://kemono.party/{service}/user/{id}
"""

import re
from urllib.parse import urljoin

from ..core import Episode

domain = ["kemono.party"]
name = "Kemono"
noepfolder = True

def get_title(html, url):
	service = re.search('<meta name="service" content="([^"]+)', html).group(1) # type: ignore
	name = re.search('<meta name="artist_name" content="([^"]+)', html).group(1) # type: ignore
	return f"[Kemono][{service}] {name}"

def get_episodes(html, url):
	result = []
	for match in re.finditer(r'<a href="([^"]+/post/(\d+))">\s*<header[^>]*>([^<]*)</header>', html):
		id = match.group(2).strip()
		title = match.group(3).strip()
		result.append(Episode(
			title=f"{id} - {title}",
			url=urljoin(url, match.group(1))
		))
	result.reverse()
	return result

def get_images(html, url):
	result = []
	for match in re.finditer(r'<a[^>]*href="([^"]*)"\s+download', html):
		result.append(match.group(1))
	return result

def get_next_page(html, url):
	if "/post/" in url:
		return None
	match = re.search(r'<a href="([^"]+)"[^>]*class="next"', html)
	if match:
		return urljoin(url, match.group(1))



#! python3

import re
from html import unescape
from urllib.parse import urlparse, parse_qs, quote, urljoin

from ..core import Episode
from ..error import PauseDownloadError, SkipEpisodeError

domain = ["chan.sankakucomplex.com"]
name = "Sankaku"
noepfolder = True
config = {
	# curl for chan.sankakucomplex.com
	"curl": "",
	# curl for v.sankakucomplex.com. Note that you should leave this empty.
	"curl_v": ""
}
autocurl = True

def login_check(html):
	if '<a href="/user/login">' in html:
		raise PauseDownloadError("You didn't login")

def get_title(html, url):
	title = re.search(r"<title>/?(.+?) \|", html).group(1)
	return "[sankaku] " + title
	
next_page_cache = {}

def get_episodes(html, url):
	login_check(html)
	s = []
	pid = None
	for m in re.finditer(r'href="(/(?:[^/]*/)?post/show/(\d+))"', html):
		ep_url, pid = m.groups()
		e = Episode(pid, urljoin(url, ep_url))
		s.append(e)
	
	if len(s) > 1:
		# breakpoint()
		tags = parse_qs(urlparse(url).query)["tags"][0]
		tags = quote(tags)
		next_page_cache[url] = f"https://chan.sankakucomplex.com/?tags={tags}&next={pid}"
		
	return s[::-1]

def get_images(html, url):
	if "This post was deleted" in html:
		raise SkipEpisodeError(always=True)
	login_check(html)
	result = ""
	if match := re.search('<a [^>]*highres[^>]*>', html):
		result = re.search('href="([^"]+)"', match.group(0)).group(1)
	elif match := re.search('embed src="([^"]+)"', html):
		result = match.group(1)
	return [urljoin(url, unescape(result))]

def get_next_page(html, url):
	if url in next_page_cache:
		return next_page_cache.pop(url)
		



#! python3

"""this is Wix module for comiccrawler
"""

import re
import json

from ..core import Episode, grabhtml

domain = ["wix.com"]
name = "Wix"
noepfolder = True

def get_title(html, url):
	return "[Wix.com] " + re.search("<title>([^<]+)", html).group(1)
	
def trim_ext(text):
	return re.sub("\.(gif|png|jpg)$", "", text, flags=re.I).strip()
	
def get_episodes(html, url):
	public_model = re.search("var publicModel = ({.+})", html).group(1)
	public_model = json.loads(public_model)
	
	pages = public_model["pageList"]["pages"]
	
	s = []
	
	for page in pages:
		try:
			data = grabhtml(page["urls"][0])
		except KeyError:
			data = grabhtml("https://static.wixstatic.com/sites/" + page["pageJsonFileName"] + ".z?v=3")
		
		data = json.loads(data)
		data = data["data"]["document_data"]
		
		for item in data.values():
			if item["type"] != "Image":
				continue
			s.append(Episode(
				"{} - {}".format(page["title"], trim_ext(item.get("title", "")) or item["id"])		,
				"https://static.wixstatic.com/media/" + item["uri"],
				image="https://static.wixstatic.com/media/" + item["uri"]
			))
			
	return s



#! python3
"""
http://www.dmzj.com/info/xixingji.html

"""

import re
from urllib.parse import urljoin

from deno_vm import eval

from ..core import Episode

domain = ["www.dmzj.com"]
name = "動漫之家w"

def get_title(html, url):
	return re.search("comic_name = '([^']+)", html).group(1)

def get_episodes(html, url):
	# dmzj repeat the list twice?...
	end = html.index('<div class="fg"></div>')
	html = html[:end]

	view_url = re.escape(urljoin(url, "/view"))
	pattern = 'href="({view_url}[^"]+)[^>]+?title="([^"]+)'.format(view_url=view_url)
	s = []
	for match in re.finditer(pattern, html):
		ep_url, title = match.groups()
		s.append(Episode(title, ep_url))
	s = s[::-1]
	
	return s

def get_images(html, url):
	# Set base url
	base = "http://images.dmzj.com/"

	# Get urls
	s = re.search(r"page = '';\s+([^\n]+)", html).group(1)
	pages = eval(s + "; pages")
	pages = re.search('"page_url":"([^"]+)', pages).group(1)
	pages = re.split("\r?\n", pages)

	# thumbs.db?!
	# http://manhua.dmzj.com/zhuoyandexiana/3488-20.shtml
	return [base + page for page in pages if page and not page.lower().endswith("thumbs.db")]



#! python3

"""this is konachan module for comiccrawler

Ex:
	http://konachan.com/pool/show/218

"""

import re
from html import unescape
from urllib.parse import urljoin

from ..core import Episode
from ..error import SkipEpisodeError

domain = ["konachan.com"]
name = "Konachan"
noepfolder = True

def get_title(html, url):
	title = re.search(r"<title>/?(.+?) \|", html).group(1)
	return "[konachan] " + title
	
def get_episodes(html, url):
	s = []
	base = re.search("(https?://[^/]+)", url).group(1)
	for m in re.finditer('<a class="thumb" href="([^"]+)"', html):
		url = m.group(1)
		uid = re.search(r"show/(\d+)", url).group(1)
		e = Episode(uid, base + url)
		s.append(e)
	return s[::-1]

def get_images(html, url):
	try:
		img = re.search('href="([^"]+)" id="highres"', html).group(1)
	except AttributeError:
		if "This post was deleted" in html:
			raise SkipEpisodeError from None
		raise
	return urljoin(url, img)

def get_next_page(html, url):
	match = re.search('<a class="next_page" rel="next" href="([^"]+)">', html)
	if match:
		return urljoin(url, unescape(match.group(1)))
		


#! python3

"""this is buka module for comiccrawler

Ex:
	http://www.buka.cn/detail/202796

"""

import re

from urllib.parse import urljoin

from ..core import Episode

domain = ["www.buka.cn"]
name = "布卡"

def get_title(html, url):
	return re.search(r'class="title-font">([^<]*)', html).group(1).strip()

def get_episodes(html, url):
	rx = '<a[^>]+href="(/view/[^"]+)"[^>]*>([^<]+)'
	arr = []
	for match in re.finditer(rx, html):
		ep_url, title = match.groups()
		title = title.strip()
		arr.append(Episode(title, urljoin(url, ep_url)))
	return arr[::-1]
	
def get_images(html, url):
	return re.findall('="([^"]+?/pics/[^"]+?)"', html)



#! python3

"""this is facebook module for comiccrawler

Ex:
	https://www.facebook.com/gtn.moe/photos/pcb.942148229165656/942147902499022/?type=3&theater
	https://www.facebook.com/Nisinsheep/photos/pcb.1667254973516031/1667254333516095/?type=3&theater
	https://www.facebook.com/photo.php?fbid=10202533447211476&set=a.10202533447131474.1073741835.1654599523&type=3&theater
	https://www.facebook.com/photo.php?fbid=10210069591680987&set=p.10210069591680987&type=3&theater

"""

import re, json
from html import unescape
from ..core import Episode

domain = ["www.facebook.com"]
name = "FB"
circular = True
noepfolder = True
config = {
	"curl": ""
}
autocurl = True

def get_title(html, url):
	try:
		id = re.search(r"photos/([^/]+)", url).group(1)
	except AttributeError:
		id = re.search("set=([^&]+)", url).group(1)
	title = re.search("<title[^>]*>([^<]+)", html).group(1)
	title = re.sub("\s+", " ", title)
	return unescape("{} ({})".format(title, id))

def get_episodes(html, url):
	return [Episode("image", url)]
	
def get_images(html, url):
	# with open("test.html", "w", encoding="utf-8") as f:
	# 	f.write(html)
	i = html.index("currMedia")
	rx = re.compile(r'"image":\{"uri":("[^"]+")')
	uri = rx.search(html, i).group(1)
	uri = json.loads(uri)
	return [uri]

def get_next_image_page(html, url):
	i = html.index("nextMediaAfterNodeId")
	rx = re.compile(r'"id":"([^"]+)"')
	next_id = rx.search(html, i).group(1)
	return get_next_url(url, next_id)

def get_next_url(url, next_id):
	match = re.match(r'(https://www\.facebook\.com/[^/]+/photos/[^/]+/)\d+(.*)', url)
	if match:
		return match.group(1) + next_id + match.group(2)

	match = re.match(r'(.*fbid=)\d+(.*)', url)
	if match:
		return match.group(1) + next_id + match.group(2)

	raise TypeError("Can't find next url")



#! python3

"""this is exh module for comiccrawler, support exhentai.org, g.e-hentai.org

"""

import re

from html import unescape
from urllib.parse import urljoin
from configparser import ConfigParser

from ..episode import Episode
from ..image import Image
from ..util import url_extract_filename
from ..error import PauseDownloadError

domain = ["exhentai.org", "e-hentai.org"]
name = "e紳士"
noepfolder = True
rest = 5
config = {
	"cookie_ipb_member_id": "請輸入Cookie中的ipb_member_id",
	"cookie_ipb_pass_hash": "請輸入Cookie中的ipb_pass_hash",
	"cookie_igneous": "請輸入Cookie中的igneous",
	"original": "false"
}

def get_boolean(s):
	return ConfigParser.BOOLEAN_STATES.get(s.lower())
	
def check_login(html):
	if html[6:10] == "JFIF":
		raise PauseDownloadError("You didn't login!")

def get_title(html, url):
	check_login(html)
	title = re.findall("<h1 id=\"g(j|n)\">(.+?)</h1>", html)[-1][1]
	hash = re.search(r"/g/(\w+/\w+)", url).group(1)
	return unescape("{title} ({hash})".format(title=title, hash=hash))

def get_episodes(html, url):
	check_login(html)
	url = re.search(r'href="([^\"]+?/s/\w+/\w+-1)"', html).group(1)
	e = Episode("image", url)
	return [e]

nl = ""
def get_images(html, url):
	check_login(html)
	global nl
	nl = re.search("nl\('([^)]+)'\)", html).group(1)
	
	image = re.search("<img id=\"img\" src=\"(.+?)\"", html)
	image = unescape(image.group(1))
	# bandwith limit
	if re.search("509s?\.gif", image) or re.search("403s?\.gif", image):
		# pause the download since retry doesn't help but increase view limit.
		raise PauseDownloadError("Bandwidth limit exceeded!")
		
	filename = url_extract_filename(image)

	if get_boolean(config["original"]):
		match = re.search(r'href="([^"]+?/fullimg\.php[^"]+)', html)
		if match:
			image = unescape(match.group(1))

	return Image(image, filename=filename)

def errorhandler(er, crawler):
	url = urljoin(crawler.ep.current_url, "?nl=" + nl)
	if crawler.ep.current_url == url:
		crawler.ep.current_url = url.split("?")[0]
	else:
		crawler.ep.current_url = url
	crawler.html = None

def get_next_page(html, url):
	check_login(html)
	match = re.search('id="next"[^>]+?href="([^"]+)', html)
	if match:
		match = match.group(1)
		if match not in url:
			return match



#! python3

import re
from html import unescape

from ..core import Episode
from ..url import urljoin
from ..util import clean_tags

domain = ["nijie.info"]
name = "nijie"
noepfolder = True
config = {
	"cookie_n_session_hash": "",
	"cookie_nemail": "",
	"cookie_nlogin": ""
}

def get_title(html, url):
	match = re.search(
		r'<a class="name" href="members\.php\?id=(\d+)">(.+?)</a>', html)
	uid, name = match.groups()
	
	name = clean_tags(name)
	
	if "dojin.php" in url:
		name += " (dojin)"
		
	return "[nijie] {id} - {name}".format(id=uid, name=name)

def get_episodes(html, url):
	s = []
	ep_set = set()
	
	for m in re.finditer(r'<a href="(/view\.php\?id=(\d+))" title="([^"]+)', html):
		ep_url, ep_id, title = m.groups()
		if ep_id in ep_set:
			continue
		ep_set.add(ep_id)
		
		title = "{} - {}".format(ep_id, unescape(title))
		ep_url = urljoin(url, ep_url)
		
		s.append(Episode(title, ep_url))
		
	return s[::-1]
	
def clean_rs(url):
	return re.sub("/__rs_[^/]+", "", url)
	
def get_images(html, url):
	try:
		html = html[html.index('id="view-center"'):]	
	except ValueError:
		pass
	html = html[:html.index('id="nuitahito')]
	
	matches = re.finditer(r'<img[^>]+src="([^"]+)[^>]+data-original', html)
	matches = list(matches)
	if not matches:
		matches = re.finditer(r'<img[^>]+?illust_id=[^>]+?src="([^"]+)', html)
	return list(urljoin(url, clean_rs(m.group(1))) for m in matches)

def get_next_page(html, url):
	match = re.search(r'<a rel="next" href="([^"]+)', html)
	if match:
		return urljoin(url, match.group(1))



#! python3

"""動漫狂"""

import re

from urllib.parse import urljoin

from ..core import Episode

domain = ["www.cartoonmad.com"]
name = "動漫狂"

def get_title(html, url):
	return re.search("<title>([^<]+) - [^<]+? - [^<]+<", html).group(1)

def get_episodes(html, url):
	s = []
	for match in re.finditer('href=(/comic/\d{6,}[^\s>]+).*?>([^<]+)', html):
		ep_url, title = match.groups()
		s.append(Episode(title, urljoin(url, ep_url)))
	return s

def get_images(html, url):
	pic_url = re.search('img src="([^"]*comicpic[^"]+)', html).group(1)
	return urljoin(url, pic_url)
	
def get_next_page(html, url):
	match = re.search('a href="(\d+[^"]*)', html)
	if match:
		return urljoin(url, match.group(1))



#! python3

"""oh 漫畫

https://www.ohmanhua.com/13336/
"""

import re
from html import unescape
from urllib.parse import urljoin

from deno_vm import eval

from ..core import Episode, grabhtml

domain = ["www.ohmanhua.com", "www.cocomanhua.com"]
name = "OH漫畫"

def get_title(html, url):
	title = re.search('meta property="og:comic:book_name" content="([^"]+)', html).group(1)
	return unescape(title).strip()

def get_episodes(html, url):
	s = []
	for match in re.finditer(r'title="([^"]+)" href="([^"]+)', html):
		title, ep_url = match.groups()
		s.append(Episode(
			unescape(title),
			urljoin(url, ep_url)
		))
	return s[::-1]

class ScriptCache:
	def __init__(self):
		self.cache = {}
		
	def fetch(self, html, url, scripts):
		for script in scripts:
			if script in self.cache:
				continue
			pattern = 'src="([^"]+{})'.format(script)
			js_url = re.search(pattern, html).group(1)
			self.cache[script] = grabhtml(urljoin(url, js_url))
		
	def __str__(self):
		return "\n".join(self.cache.values())
	
scripts = ScriptCache()

def get_images(html, url):
	cdata = re.search("var C_DATA='[^']+'", html).group(0)
	
	scripts.fetch(html, url, [
		"\/l\.js",
		"common\.js",
		"custom\.js",
		"manga\.read\.js"
	])
	
	code = """
	(function() {
	
	function noop(path = "") {
	  if (path === "document.cookie") return "";
	  if (path === "$.inArray") return (v, a) => a.indexOf(v);
	  
	  return new Proxy(() => {}, {
		apply: () => noop("?"),
		get: (target, prop) => noop(`${path}.${prop}`)
	  });
	}
	
	const exports = undefined;
	const window = global;
	window.location = {
    	protocol: "http://",
		href: '""" + url + """'
	}
	const navigator = {
	  userAgent: ""
	};
	const document = noop("document")
	const $ = noop("$");
	
	""" + cdata + "\n" + str(scripts) + """
	
	window.use_domain = {
	},
	window.lines = {
	  [mh_info.chapter_id]: {
		use_line: mh_info.domain
	  }
	};
	window.chapter_id = mh_info.chapter_id;
	
	const imgs = [];
	let dirty = false;
	class Image {
		set src(val) {
			imgs.push(val);
			dirty = true;
		}
	}
	
	let i = mh_info.startimg;
	do {
		dirty = false;
		__cr.preLoadImg(i++)
	} while (dirty);
	return imgs;
	}).call(global);
	"""
	
	imgs = eval(code)
	return [urljoin(url, i) for i in imgs]



from .util import url_extract_filename

class Image:
	"""Image container"""
	def __init__(self, url=None, get_url=None, data=None, filename=None, ext=None, static_filename=None):
		self.url = url
		self.get_url = get_url
		self.data = data
		self.filename = filename
		if static_filename is None:
			self.static_filename = bool(filename)
		else:
			self.static_filename = static_filename
		self.ext = ext
		
		if not filename and url:
			self.filename = url_extract_filename(url)
			
	def resolve(self):
		if not self.url and self.get_url:
			self.url = self.get_url()
		
		if not self.filename and self.url:
			self.filename = url_extract_filename(self.url)
	
	@classmethod
	def create(cls, data):
		if isinstance(data, Image):
			return data
			
		if isinstance(data, str):
			return Image(url=data)
			
		if callable(data):
			return Image(get_url=data)

		if isinstance(data, dict):
			return Image(**data)
			
		return Image(data=data)



#! python3

import re
from urllib.parse import urljoin
from ..core import Episode, grabhtml

domain = ["www.99comic.com", "99.hhxxee.com"]
name = "99"

def get_title(html, url):
	return re.search("<h1><a title='([^']+)'", html).group(1)
	
def get_episodes(html, url):
	s = []
	# base = re.search("(https?://[^/]+)", url).group(1)
	for m in re.finditer("href='(/comics/[^']+/)'>([^<]+)</a>(?!</li>)", html):
		ep_url, title = m.groups()
		s.append(Episode(title, urljoin(url, ep_url)))
	return s[::-1]

def get_images(html, url):
	viewhtml_js = re.search('src="([^"]+viewhtml\.js)"', html).group(1)
	viewhtml_js = grabhtml(urljoin(url, viewhtml_js))
	
	ds = re.search('sDS = "([^"]+)', viewhtml_js).group(1)
	ds = ds.split("|")
	
	sfiles = re.search("sFiles=\"([^\"]+)\"", html).group(1)
	spath = re.search("sPath=\"(\d+)\"", html).group(1)
	
	imgs = sfiles.split("|")
	base = ds[int(spath) - 1]
	
	return [base + i for i in imgs]



from threading import Lock

from .error import ModuleError
from .channel import mission_ch
from .mods import get_module

class Mission:
	"""Create Mission object. Contains information of the mission."""

	def __init__(self, title=None, url=None, episodes=None, state="INIT", last_update=None):
		"""Construct."""
		super().__init__()

		self.title = title
		self.url = url
		self.episodes = episodes
		self.state = state
		self.module = get_module(url)
		self.last_update = last_update
		if not self.module:
			raise ModuleError("Get module failed!")
			
mission_lock = Lock()
class MissionProxy:
	"""Publish MISSION_PROPERTY_CHANGED event when property changed"""
	def __init__(self, mission):
		self.__dict__["mission"] = mission

	def __getattr__(self, name):
		return getattr(self.mission, name)

	def __setattr__(self, name, value):
		with mission_lock:
			setattr(self.mission, name, value)
		mission_ch.pub("MISSION_PROPERTY_CHANGED", self)

	def tojson(self):
		json = vars(self.mission).copy()
		del json["module"]
		return json

def create_mission(*args, **kwargs):
	return MissionProxy(Mission(*args, **kwargs))



#! python3
import webbrowser
import tkinter as tk
from tkinter import ttk

from belfrywidgets import ToolTip

from .core import safe_tk
from .dialog import Dialog

class SelectEpisodeDialog(Dialog):
	def __init__(self, parent, title=None, mission=None, on_closed=None):
		self.mission = mission
		
		self.checks = []
		self.anchor_index = None
		
		self.canvas = None
		self.window = None
		self.window_column = 0
		self.window_left = 0
		self.style = None
		
		super().__init__(parent, title, on_closed=on_closed)
		
	def create_window(self, column):
		if self.window:
			self.window.update_idletasks()
			self.window_left += self.window.winfo_reqwidth()
			self.window_column = column
			
		self.window = ttk.Frame(self.canvas)
		self.canvas.create_window((self.window_left, 0), window=self.window,
			anchor="nw")
			
	def create_checkbutton(self, ep, index):
		# click handler for range selection
		def handle_click(event):
			if event.state & 0x0001 and self.anchor_index is not None: # shift
				start = min(self.anchor_index, index)
				end = max(self.anchor_index, index)
				for i in range(start, end + 1):
					if i in (index, self.anchor_index):
						continue
					check = self.checks[i]
					if check.instate(("selected", )):
						check.state(("!selected", ))
					else:
						check.state(("selected", ))
			else:
				self.anchor_index = index

		def handle_right_click(_event):
			webbrowser.open(ep.url)
				
		style = "TCheckbutton"
		if ep.complete:
			style = "Complete.TCheckbutton"
		check = ttk.Checkbutton(self.window, text=safe_tk(ep.title), style=style)
		check.bind("<ButtonRelease-1>", handle_click)
		check.bind("<ButtonRelease-3>", handle_right_click)
		check.state(("!alternate",))
		if not ep.skip:
			check.state(("selected",))
		check.grid(
			column=(index // 20) - self.window_column,
			row=index % 20,
			sticky="w"
		)
		self.checks.append(check)
		
	def create_column_checkbutton(self, index):
		def check_col():
			start_index = index // 20 * 20
			if check.instate(("selected",)):
				state = ("selected",)
			else:
				state = ("!selected",)
			for ep_check in self.checks[start_index:index + 1]:
				ep_check.state(state)
				
		check = ttk.Checkbutton(self.window, command=check_col)
		state = ("!alternate",)
		if self.mission.module.config.getboolean("selectall"):
			state += ("selected",)
		check.state(state)
		check.grid(
			column=(index // 20) - self.window_column,
			row=20,
			sticky="we"
		)

	def create_body(self):
		self.style = ttk.Style()
		self.style.configure("Complete.TCheckbutton", foreground="green")

		xscrollbar = ttk.Scrollbar(self.body, orient="horizontal")
		self.canvas = tk.Canvas(
			self.body,
			xscrollcommand=xscrollbar.set,
			highlightthickness="0"
		)

		for i, ep in enumerate(self.mission.episodes):
			# create a new window for every 200 items
			if i % 200 == 0:
				self.create_window(i // 20)
			
			# create checkbutton
			self.create_checkbutton(ep, i)
			
			# checkbutton for each column
			if i % 20 == 19 or i == len(self.mission.episodes) - 1:
				self.create_column_checkbutton(i)
				
		# Resize canvas
		self.canvas.update_idletasks()
		cord = self.canvas.bbox("all")
		self.canvas.config(
			scrollregion=cord,
			height=cord[3],
			width=cord[2]
		)

		# caculates canvas's size then deside whether to show scrollbar
		def decide_scrollbar(_event):
			if self.canvas.winfo_width() >= self.canvas.winfo_reqwidth():
				xscrollbar.pack_forget()
				self.canvas.unbind("<Configure>")
		self.canvas.bind("<Configure>", decide_scrollbar)

		# draw innerframe on canvas then show
		self.canvas.pack()

		# link scrollbar to canvas then show
		xscrollbar.config(command=self.canvas.xview)
		xscrollbar.pack(fill="x")
		
	def create_buttons(self):
		btn = ttk.Button(self.btn_bar, text="反相", command=self.toggle)
		btn.pack(side="left")
		ToolTip(btn, "點擊任意項目後，Shift + 點擊另一項目，會對其間的項目反相")
		super().create_buttons()

	def apply(self):
		count = 0
		for ep, ck in zip(self.mission.episodes, self.checks):
			ep.skip = not ck.instate(("selected",))
			count += not ep.skip
		return count

	def toggle(self):
		for ck in self.checks:
			if ck.instate(("selected", )):
				ck.state(("!selected", ))
			else:
				ck.state(("selected", ))
		
def select_episodes(parent, mission, on_closed=None):
	"""Create dialog to select episodes."""
	dialog = SelectEpisodeDialog(
		parent,
		title="選擇集數",
		mission=mission,
		on_closed=on_closed
	)
	if on_closed:
		return None
	return dialog.wait()



#! python3

"""twitter"""

import re
import json

from collections import OrderedDict
from urllib.parse import urlparse, parse_qs

from ..episode import Episode
from ..grabber import grabber
from ..url import update_qs
from ..util import extract_curl
from ..error import is_http, SkipEpisodeError

domain = ["twitter.com"]
name = "twitter"
config = {
	"curl": ""
}
noepfolder = True

next_page_cache = {}
pin_entry_cache = {}

def get_title(html, url):
	name = re.search("twitter.com/([^/]+)", url).group(1)
	if is_media(url):
		return f"[twitter] {name} (media)"
	return f"[twitter] {name}"
	
def grabhandler(grab_method, url, **kwargs):
	if url.startswith("https://twitter.com/i/api/"):
		return grab_json(url, **kwargs)
		
def grab_json(url, **kwargs):
	_url, header, cookie = extract_curl(config["curl"])
	# NOTE: method-level header/cookies won't be stored into session
	kwargs["headers"] = header
	kwargs["cookies"] = cookie
	return grabber(url, **kwargs).json()
	
def is_media(url):
	if re.search(r"twitter\.com/[^/]+/media", url):
		return True
	if re.search(r"graphql/[^/]+/UserMedia", url):
		return True
	return False

def get_episodes(html, url):
	name = re.search(r"twitter\.com/([^/]+)", url).group(1)
	if name != "i":
		variables = {
			"screen_name": name,
			"withSafetyModeUserFields": True,
			"withSuperFollowsUserFields": False
		}
		u = update_qs(
			"https://twitter.com/i/api/graphql/LPilCJ5f-bs3MjJJNcuuOw/UserByScreenName",
			{"variables": json.dumps(variables)}
		)
		result = grab_json(u)
		uid = result["data"]["user"]["result"]["rest_id"]
		
		endpoint = user_media_graph if is_media(url) else user_tweets_graph
		next_page_cache[url] = endpoint(userId=uid)
		return
		
	if any(k in url for k in ["UserTweets", "UserMedia"]):
		instructions = html["data"]["user"]["result"]["timeline"]["timeline"]["instructions"]
		for instruction in reversed(instructions):
			# FIXME: we have to process pin entry first
			if instruction["type"] == "TimelinePinEntry":
				extract_pin_entry(instruction["entry"], url)
			
			if instruction["type"] == "TimelineAddEntries":
				yield from reversed(list(extract_added_entries(instruction["entries"], url)))
				
def tweet_result_to_episode(tweet_result):
	try:
		all_media = tweet_result["legacy"]["entities"]["media"] + tweet_result["legacy"]["extended_entities"]["media"]
	except KeyError:
		return None
	imgs = [find_media_source(m) for m in all_media]
	imgs = list(OrderedDict.fromkeys(imgs).keys()) # remove dup
	result = None
	try:
		result = tweet_result["legacy"]["retweeted_status_result"]["result"]
	except KeyError:
		result = tweet_result
	
	screen_name = result["core"]["user_results"]["result"]["legacy"]["screen_name"]
	id_str = result["legacy"]["id_str"]	
	ep_url = f"https://twitter.com/{screen_name}/status/{id_str}"
	
	return Episode(
		title=id_str,
		url=ep_url,
		image=imgs
	)

def extract_pin_entry(entry, url):
	ep = tweet_result_to_episode(entry["content"]["itemContent"]["tweet_results"]["result"])
	if not ep:
		return
	user_id = parse_graph_variable(url)["userId"]
	pin_entry_cache[(user_id, is_media(url))] = ep

def extract_added_entries(entries, url):
	user_id = parse_graph_variable(url)["userId"]
	pin_entry_key = (user_id, is_media(url))
	pinned_entry = pin_entry_cache.get(pin_entry_key)
	cursor = None
	has_timeline = False
	
	for entry in entries:
		if entry["content"]["entryType"] == "TimelineTimelineItem":
			has_timeline = True
			ep = tweet_result_to_episode(entry["content"]["itemContent"]["tweet_results"]["result"])
			if not ep:
				continue
				
			if pinned_entry and url_to_id(pinned_entry.url) > url_to_id(ep.url):
				yield pinned_entry
				del pin_entry_cache[pin_entry_key]
				pinned_entry = None
				
			yield ep
			
		if entry["content"]["entryType"] == "TimelineTimelineCursor" and entry["content"]["cursorType"] == "Bottom":
			cursor = entry["content"]["value"]
				
	if cursor and has_timeline:
		endpoint = user_media_graph if is_media(url) else user_tweets_graph
		next_page_cache[url] = endpoint(userId=user_id, cursor=cursor)
		
	if not has_timeline and pinned_entry:
		yield pinned_entry
		del pin_entry_cache[pin_entry_key]

def url_to_id(url):
	tid = re.search(r"\d+$", url).group()
	return int(tid)
	
def find_media_source(media):
	if media["type"] == "video":
		max_video = sorted(media["video_info"]["variants"], key=lambda i: i.get("bitrate", 0)).pop()
		return max_video["url"]
	return media["media_url_https"] + ":orig"
	
def parse_graph_variable(url):
	variables = parse_qs(urlparse(url).query)["variables"][0]
	return json.loads(variables)
	
def user_tweets_graph(**kwargs):
	variables = {
		"count": 20,
		"withTweetQuoteCount": True,
		"includePromotedContent": True,
		"withSuperFollowsUserFields": False,
		"withUserResults": True,
		"withBirdwatchPivots": False,
		"withReactionsMetadata": False,
		"withReactionsPerspective": False,
		"withSuperFollowsTweetFields": False,
		"withVoice": True
	}
	variables.update(kwargs)
	return update_qs(
		"https://twitter.com/i/api/graphql/PIt4K9PnUM5DP9KW_rAr0Q/UserTweets",
		{"variables": json.dumps(variables)}
	)
	
def user_media_graph(**kwargs):
	variables = {
		"userId": "1240968347375656960",
		"count": 20,
		"withHighlightedLabel": True,
		"withTweetQuoteCount": False,
		"includePromotedContent": False,
		"withSuperFollowsUserFields": False,
		"withUserResults": True,
		"withBirdwatchPivots": False,
		"withReactionsMetadata": True,
		"withReactionsPerspective": True,
		"withSuperFollowsTweetFields": False,
		"withClientEventToken": False,
		"withBirdwatchNotes": False,
		"withVoice": True
	}
	variables.update(kwargs)
	return update_qs(
		"https://twitter.com/i/api/graphql/JWaFyG5p4-UvSyxGMe15-g/UserMedia",
		{"variables": json.dumps(variables)}
	)

def tweet_detail_graph(**kwargs):
	variables = {"focalTweetId":"1438138335042564102","with_rux_injections":False,"includePromotedContent":True,"withCommunity":True,"withQuickPromoteEligibilityTweetFields":True,"withTweetQuoteCount":True,"withBirdwatchNotes":False,"withSuperFollowsUserFields":True,"withBirdwatchPivots":False,"withDownvotePerspective":False,"withReactionsMetadata":False,"withReactionsPerspective":False,"withSuperFollowsTweetFields":True,"withVoice":True,"withV2Timeline":False,"__fs_interactive_text":False,"__fs_dont_mention_me_view_api_enabled":False}
	variables.update(kwargs)
	return update_qs(
		"https://twitter.com/i/api/graphql/s2RO46g9Rhw53GX2BEMfiA/TweetDetail",
		{"variables": json.dumps(variables)}
	)

def errorhandler(err, crawler):
	if is_http(err, 404):
		tid = re.search("status/(\d+)", crawler.ep.current_url).group(1)
		u = tweet_detail_graph(focalTweetId=tid)
		r = grab_json(u)
		if r.get("errors", None):
			if "No status found" in r["errors"][0]["message"]:
				raise SkipEpisodeError(always=True)

def get_next_page(html, url):
	return next_page_cache.pop(url, None)



#! python3

"""module for senmanga
http://raw.senmanga.com/Overlord/
"""

import re

from ..core import Episode
from ..url import urljoin

domain = ["raw.senmanga.com"]
name = "senmanga"

def get_title(html, url):
	return re.search(r'<h1 class="title"><a[^>]*>([^<]+)', html).group(1)

def get_episodes(html, url):
	prefix = re.escape(url)
	s = []
	for m in re.finditer(r'<a href="({}/[^/"]+/[^"]+)"[^>]*>([^<]+)'.format(prefix), html):
		ep_url, title = m.groups()
		s.append(Episode(title, urljoin(url, ep_url)))
	return s[::-1]

def get_images(html, url):
	return re.search(r'<img src="([^"]+?/viewer/[^"]+)', html).group(1)
	
def get_next_page(html, url):
	match = re.search('<a href="([^"]+)"><[^>]+>Next Page</span>', html)
	if match:
		next_page_url = urljoin(url, match.group(1))
		if get_ep_from_url(url) == get_ep_from_url(next_page_url):
			return next_page_url

def get_ep_from_url(url):
	match = re.match("https://raw\.senmanga\.com/([^/]+)/([^/]+)(?:/([^/]+))?", url)
	_name, ep, _page = match.groups()
	return ep



#! python3

"""
https://pixabay.com/users/kellepics-4893063/
"""

import re
from html import unescape
from urllib.parse import urljoin

from ..episode import Episode

domain = ["pixabay.com"]
name = "pixabay"
noepfolder = True
config = {
	"cookie_sessionid": ""
}

def get_title(html, url):
	title = re.search("<h1[^>]*>([^<]+)", html).group(1)
	return "[pixabay] {}".format(unescape(title))
	
def get_episodes(html, url):
	pattern = '<a href="(/\w+/[\w-]+-(\d+)/)".+?(?:src|data-lazy)="[^"]+/([\w-]+\d+)__\d+(\.[\w]+)" alt="([^"]+)'
	s = []
	for match in re.finditer(pattern, html):
		ep_url, id, name, ext, alt = match.groups()
		title = "{} - {}".format(id, alt)
		ep = Episode(
			title=title,
			url=urljoin(url, ep_url),
			image="https://pixabay.com/images/download/{}{}?attachment".format(name, ext)
		)
		s.append(ep)
	return s[::-1]

def get_next_page(html, url):
	match = re.search('<a class="[^"]*?next" href="([^"]+)', html)
	if match:
		return urljoin(url, match.group(1))



#! python3

"""this is deviant module for comiccrawler, support deviantart.com

2014/7/13
JDownloader2 now support　deviantart.com!

"""

import re
import json
from html import unescape
from urllib.parse import urlparse

from deno_vm import eval

from ..error import PauseDownloadError
from ..core import Episode
from ..url import update_qs

domain = ["deviantart.com"]
name = "dA"
noepfolder = True
config = {
	"cookie_auth": "",
	"cookie_auth_secure": "",
	"cookie_userinfo": ""
}

def get_title(html, url):
	id = urlparse(url).path[1:]
	return f"[dA] {id}"
	
def check_login(state):
	if not state["@@publicSession"]["isLoggedIn"]:
		raise PauseDownloadError("You didn't login")		
		
next_page_cache = {}
		
def get_episodes_from_api(source, url):
	data = json.loads(source)
	s = []
	for item in data["results"]:
		d = item["deviation"]
		s.append(Episode(
			f"{d['deviationId']} - {d['title']}",
			d["url"]
		))
		
	if data["hasMore"]:
		next_page_cache[url] = update_qs(url, {
			"offset": data["nextOffset"]
		})
		
	return s[::-1]
	
def get_state(html):
	js = re.search(r"(window\.__INITIAL_STATE__[\s\S]+?)</script>", html).group(1)
	return eval("""
		const window = {};
	""" + js + """
		; window.__INITIAL_STATE__
	""")	

def get_episodes(html, url):
	if "_napi" in url:
		return get_episodes_from_api(html, url)
		
	state = get_state(html)
	check_login(state)

	deviation = state["@@entities"]["deviation"]
	gallery = None
	for key, value in state["@@streams"].items():
		if key.startswith("folder-deviations-gallery"):
			gallery = value
			break
	s = []
	for id in gallery["items"]:
		d = deviation[str(id)]
		s.append(Episode(
			f"{id} - {d['title']}",
			d["url"]
		))
		
	if gallery["hasMore"]:
		username = next(iter(state["@@entities"]["user"].values()))["username"]
		params = {
			"username": username,
			"offset": gallery["nextOffset"],
			"limit": gallery["itemsPerFetch"],
			# FIXME: it seems that this param is not used with featured galleries?
			"mode": "newest"
		}
		folder_id = gallery["streamParams"]["folderId"]
		if folder_id < 0:
			params["all_folder"] = "true"
		else:
			params["folderid"] = folder_id
		next_page_cache[url] = update_qs(
			"https://www.deviantart.com/_napi/da-user-profile/api/gallery/contents",
			params)
		
	return s[::-1]

def get_images(html, url):
	state = get_state(html)
	check_login(state)

	try:
		i = re.search('href="(https://www.deviantart.com/download[^"]+)', html).group(1)
		return [unescape(i)]
	except AttributeError:
		pass
		
	id = re.search(r"\d+$", url).group()
	media = state["@@entities"]["deviation"][id]["media"]
	return [f"{media['baseUri']}?token={media['token'][-1]}"]
	
def get_next_page(html, url):
	if url in next_page_cache:
		return next_page_cache.pop(url)



#! python3

"""this is dm5 module for comiccrawler

Ex:
	http://www.dm5.com/manhua-yaojingdeweiba/
	http://www.dm5.com/manhua-ribenyaoguaidaquan/

"""

import re
from urllib.parse import urljoin

from deno_vm import eval

from ..core import Episode, grabhtml
from ..util import clean_tags

cookie = {
	"isAdult": "1",
	"fastshow": "true",
	"ComicHistoryitem_zh": "ViewType=1"
}
domain = ["www.dm5.com", "tel.dm5.com", "hk.dm5.com"]
name = "動漫屋"

def get_title(html, url):
	return re.search('DM5_COMIC_MNAME="([^"]+)', html).group(1)

def get_episodes(html, url):
	s = []

	for match in re.finditer('<li>\s*<a href="(/m\d+/)"[^>]*>(.+?)</a>', html, re.DOTALL):
		# https://github.com/eight04/ComicCrawler/issues/165
		ep_url, title = match.groups()
		s.append(Episode(
			clean_tags(title),
			urljoin(url, ep_url)
		))
		
	if "DM5_COMIC_SORT=1" in html:
		return s
	return s[::-1]

def get_images(html, url):
	key = re.search(r'id="dm5_key".+?<script[^>]+?>\s*eval(.+?)</script>', html, re.DOTALL)
	
	if key:
		key = eval(key.group(1)).split(";")[1]
		key = re.search(r"=(.+)$", key).group(1)
		key = eval(key)
		
	else:
		key = ""
		
	count = int(re.search("DM5_IMAGE_COUNT=(\d+);", html).group(1))
	cid = re.search("DM5_CID=(\d+);", html).group(1)
	mid = re.search("DM5_MID=(\d+);", html).group(1)
	dt = re.search('DM5_VIEWSIGN_DT="([^"]+)', html).group(1)
	sign = re.search('DM5_VIEWSIGN="([^"]+)', html).group(1)
	
	pages = {}
	
	def grab_page(page):
		params = {
			"cid": cid,
			"page": page + 1,
			"language": 1,
			"key": key,
			"gtk": 6,
			"_cid": cid,
			"_mid": mid,
			"_dt": dt,
			"_sign": sign
		}
		fun_url = urljoin(url, "chapterfun.ashx")
		text = grabhtml(fun_url, referer=url, params=params)
		d = eval(text)
		for i, image in enumerate(d):
			pages[i + page] = image
	
	def create_page_getter(page):
		def getter():
			if page not in pages:
				grab_page(page)
			return pages[page]
		return getter
	
	return [create_page_getter(p) for p in range(count)]



#! python3

"""Download Manager"""

from collections import deque
import re
import subprocess # nosec
import shlex
import sys
import traceback
from threading import Lock

from os.path import join as path_join
from time import time

from worker import Worker, current, await_, create_worker, async_

from .analyzer import Analyzer
from .safeprint import print
from .batch_analyzer import BatchAnalyzer
from .config import setting
from .mission import create_mission
from .crawler import download
from .profile import get as profile
from .error import PauseDownloadError
from .util import safefilepath, debug_log

from .mission_manager import mission_manager
from .channel import download_ch
from .episode_loader import load_episodes

def quote(item):
	if sys.platform == "win32":
		return subprocess.list2cmdline([item])
	return shlex.quote(item)
	
class ThreadSafeSet:
	def __init__(self):
		self.lock = Lock()
		self.obj = set()
		
	def add(self, item):
		with self.lock:
			return self.obj.add(item)
			
	def remove(self, item):
		with self.lock:
			return self.obj.remove(item)
			
	def copy(self):
		with self.lock:
			return self.obj.copy()

class DownloadManager:
	"""Create a download manager used in GUI. Manage threads."""

	def __init__(self):
		"""Construct."""
		self.download_thread = None
		self.analyze_threads = ThreadSafeSet()
		self.library_thread = None
		self.library_err_count = None
		self.batch_analyzer = None
		self.continued_failure = 0
		
		thread = current()
		
		download_ch.sub(thread)
		
		@thread.listen("DOWNLOAD_ERROR")
		def _(event):
			_err, mission = event.data
			mission_manager.drop("view", mission)
			self.continued_failure += 1

		@thread.listen("DOWNLOAD_FINISHED")
		def _(event):
			"""After download, execute command."""
			self.continued_failure = 0
			
			if event.target is not self.download_thread:
				return
				
			cmd = event.data.module.config.get("runafterdownload")
			default_cmd = setting.get("runafterdownload")
			
			commands = []
			
			if cmd:
				commands.append(cmd)
				
			if default_cmd and default_cmd not in commands:
				commands.append(default_cmd)
			
			def run_command():
				for command in commands:
					target = quote(path_join(
						profile(event.data.module.config["savepath"]),
						safefilepath(event.data.title)
					))
					if "{target}" in command:
						command = command.format(target=target)
					else:
						command += " " + target
					print(f"run command: {command}")
					try:
						await_(subprocess.call, command, shell=True) # nosec
					except (OSError, subprocess.SubprocessError):
						traceback.print_exc()

			async_(run_command)
						
		@thread.listen("DOWNLOAD_FINISHED")
		@thread.listen("DOWNLOAD_ERROR")
		def _(event):
			"""After download, continue next mission"""
			if event.target is not self.download_thread:
				return
			self.download_thread = None
			if self.continued_failure >= len(mission_manager.get_all("view", lambda m: m.state != "FINISHED")):
				print(f"連續失敗 {self.continued_failure} 次，停止下載")
				return
			self.start_download(continued=True)
				
		@thread.listen("DOWNLOAD_INVALID")
		def _(event):
			"""Something bad happened"""
			if event.target is self.download_thread:
				self.download_thread = None
				print("停止下載")

	def start_download(self, continued=False):
		"""Start downloading."""
		if self.download_thread:
			return

		if not continued:
			self.continued_failure = 0
			
		mission = mission_manager.get("view", lambda m: m.state in ("ANALYZED", "PAUSE", "ERROR", "UPDATE"))
		if mission:
			print("Start download " + mission.title)
			def do_download():
				debug_log("do_download")
				with load_episodes(mission):
					download(mission, profile(mission.module.config["savepath"]))
					
			self.download_thread = Worker(do_download).start()
		else:
			print("所有任務已下載完成")

	def stop_download(self):
		"""Stop downloading."""
		if self.download_thread:
			self.download_thread.stop()
			self.download_thread = None
			print("Stop downloading")
			
	def start_analyze(self, mission, on_finished=None):
		"""Start analyzing"""
		if mission.state not in ("ANALYZE_INIT", "INIT"):
			print(
				"Invalid state to analyze: {state}".format(state=mission.state))
			return
			
		@create_worker
		def analyze_thread():
			err = None
			with load_episodes(mission):
				try:
					Analyzer(mission).analyze()
				except BaseException as _err:
					err = _err
					raise
				else:
					mission_manager.add("view", mission)
				finally:
					if on_finished:
						on_finished(err)
					self.analyze_threads.remove(analyze_thread)
		self.analyze_threads.add(analyze_thread)
		
	def start_batch_analyze(self, missions):
		"""Start batch analyze"""
		if self.batch_analyzer:
			print("Batch analyzer is already running")
			return
			
		if isinstance(missions, str):
			missions = [
				create_mission(url=m) for m in re.split("\s+", missions) if m]
		missions = deque(missions)
				
		def gen_missions():
			while missions:
				yield missions[0]
				
		def on_item_finished(err, mission):
			if not err:
				missions.popleft()
				mission_manager.add("view", mission)
				download_ch.pub("BATCH_ANALYZE_UPDATE", list(missions))
			
		def on_finished(err):
			self.batch_analyzer = None
			download_ch.pub("BATCH_ANALYZE_END", err)
			
		self.batch_analyzer = BatchAnalyzer(
			gen_missions=gen_missions(),
			on_item_finished=on_item_finished,
			on_finished=on_finished
		)
		self.batch_analyzer.start()
		
	def stop_batch_analyze(self):
		"""Stop batch analyzer"""
		if not self.batch_analyzer:
			print("No batch analyzer exists")
			return
			
		self.batch_analyzer.stop()
		
	def stop_analyze(self):
		"""Stop analyze"""
		for thread in self.analyze_threads.copy():
			thread.stop()
			self.analyze_threads.remove(thread)

	def start_check_update(self, missions=None):
		"""Start checking library update."""
		if self.library_thread:
			print("Already checking update")
			return
			
		# set mission state to ANALYZE_INIT
		setting["lastcheckupdate"] = str(time())
		
		if missions is None:
			missions = mission_manager.get_all("library", lambda m: m.state not in ("DOWNLOADING", "ANALYZING"))
			
		for mission in missions:
			mission.state = "ANALYZE_INIT"
			
		missions = set(missions)
				
		def gen_missions():
			while True:
				mission = mission_manager.get("library", lambda m: m.state in ("ANALYZE_INIT", "ERROR") and m in missions)
				if not mission:
					return
				yield mission
				
		self.library_err_count = 0
		def on_item_finished(err, mission):
			if err:
				traceback.print_exception(err)
			if mission.state == "UPDATE":
				mission_manager.lift("library", mission)
			elif mission.state == "ERROR":
				mission_manager.drop("library", mission)
				self.library_err_count += 1
				
		def stop_on_error(err):
			return self.library_err_count > 10 or isinstance(err, PauseDownloadError)
				
		def on_finished(err):
			self.library_thread = None
			if err:
				download_ch.pub("LIBRARY_CHECK_UPDATE_FAILED", err)
				print("Failed to check update")
			else:
				print("Update checking done")
				
		self.library_thread = BatchAnalyzer(
			gen_missions=gen_missions(),
			on_item_finished=on_item_finished,
			on_finished=on_finished,
			stop_on_error=stop_on_error
		)
		self.library_thread.start()
	
	def stop_check_update(self):
		"""Stop checking update"""
		if self.library_thread:
			self.library_thread.stop()
			self.library_thread = None

	def is_downloading(self):
		return self.download_thread is not None
		
download_manager = DownloadManager()



#! python3

"""this is youhui module for comiccrawler

http://m.wuyouhui.net/meishi/shisedalu/
"""

import re
from html import unescape
from urllib.parse import urlparse, urljoin

from ..core import Episode
from ..error import SkipEpisodeError
from . import xznj120

domain = ["m.wuyouhui.net"]
name = "友繪"

def get_title(html, url):
	match = re.search('<meta property="og:title" content="([^"]+)', html)
	return unescape(match.group(1))

def get_episodes(html, url):
	path = urlparse(url).path
	pattern = '<a href="({}[^"]+)[^>]*?><span>([^<]+)'.format(path)
	s = []
	for match in re.finditer(pattern, html):
		ep_url, ep_title = match.groups()
		s.append(Episode(ep_title, urljoin(url, ep_url)))
	return s[::-1]

def get_images(html, url):
	imgs = xznj120.get_images(html, url)
	if imgs:
		return imgs
	raise SkipEpisodeError
	